/*
 * Copyright The OpenTelemetry Authors
 * SPDX-License-Identifier: Apache-2.0
 */

package io.opentelemetry.sdk.trace;

import io.opentelemetry.api.trace.SpanKind;
import io.opentelemetry.api.trace.Tracer;
import java.util.concurrent.TimeUnit;
import org.openjdk.jmh.annotations.Benchmark;
import org.openjdk.jmh.annotations.Fork;
import org.openjdk.jmh.annotations.Level;
import org.openjdk.jmh.annotations.Measurement;
import org.openjdk.jmh.annotations.OutputTimeUnit;
import org.openjdk.jmh.annotations.Param;
import org.openjdk.jmh.annotations.Scope;
import org.openjdk.jmh.annotations.Setup;
import org.openjdk.jmh.annotations.State;
import org.openjdk.jmh.annotations.Threads;
import org.openjdk.jmh.annotations.Warmup;

@State(Scope.Benchmark)
public class SpanAttributeTruncateBenchmark {

  private SdkSpanBuilder sdkSpanBuilder;

  public final String shortValue = "short";
  public final String longValue = "very_long_attribute_and_then_some_more";
  public String veryLongValue;

  @Param({"10", "1000000"})
  public int maxLength;

  @Setup(Level.Trial)
  public final void setup() {
    Tracer tracer =
        SdkTracerProvider.builder()
            .setSpanLimits(SpanLimits.builder().setMaxLengthOfAttributeValues(maxLength).build())
            .build()
            .get("benchmarkTracer");
    sdkSpanBuilder =
        (SdkSpanBuilder)
            tracer
                .spanBuilder("benchmarkSpan")
                .setSpanKind(SpanKind.CLIENT)
                .setAttribute("key", "value");

    String seed = "0123456789";
    StringBuilder longString = new StringBuilder();
    while (longString.length() < 10_000_000) {
      longString.append(seed);
    }
    veryLongValue = longString.toString();
  }

  /** attributes that don't require any truncation. */
  @Benchmark
  @Threads(value = 1)
  @Fork(1)
  @Warmup(iterations = 5, time = 1)
  @Measurement(iterations = 10, time = 1)
  @OutputTimeUnit(TimeUnit.MILLISECONDS)
  public RecordEventsReadableSpan shortAttributes() {
    RecordEventsReadableSpan span = (RecordEventsReadableSpan) sdkSpanBuilder.startSpan();
    for (int i = 0; i < 10; i++) {
      span.setAttribute(String.valueOf(i), shortValue);
    }
    return span;
  }

  /** even if we truncate, result is short. */
  @Benchmark
  @Threads(value = 1)
  @Fork(1)
  @Warmup(iterations = 5, time = 1)
  @Measurement(iterations = 10, time = 1)
  @OutputTimeUnit(TimeUnit.MILLISECONDS)
  public RecordEventsReadableSpan longAttributes() {
    RecordEventsReadableSpan span = (RecordEventsReadableSpan) sdkSpanBuilder.startSpan();
    for (int i = 0; i < 10; i++) {
      span.setAttribute(String.valueOf(i), longValue);
    }
    return span;
  }

  /** have to copy very long strings. */
  @Benchmark
  @Threads(value = 1)
  @Fork(1)
  @Warmup(iterations = 5, time = 1)
  @Measurement(iterations = 10, time = 1)
  @OutputTimeUnit(TimeUnit.MILLISECONDS)
  public RecordEventsReadableSpan veryLongAttributes() {
    RecordEventsReadableSpan span = (RecordEventsReadableSpan) sdkSpanBuilder.startSpan();
    for (int i = 0; i < 10; i++) {
      span.setAttribute(String.valueOf(i), veryLongValue);
    }
    return span;
  }
}
